/*
 * strct_f.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "complex.h"
#include "cmatutil.h"
#include "lattice.h"
#include "vecmat.h"
#include "utils.h"
#include "mbconfig.h"
#include "strct_f.h"

/* extern DEBUG_INFO dbg; */
extern EPS_CTRL	epsilon;

#define SPACE_G_FILE	"mbfitspg.dat"
#define POSEPS 1e-4
#define GE 2.0886563e-2	/*  = 2 * pi * me * e / sq(h)  */
/*
 * Physical constants (CODATA 1990)
 *	h = 2.0 * pi * 1.05457266e-34 	[J s]
 *	me = 9.1093897e-31;				[kg]
 *	e = 1.60217733e-19;				[C]
 *	c = 2.99792458e+8;				[m s^{-1}]
 */


/* ================================================== */

complex fg_to_vg(complex fg, CELL_PAR bv)
{
  /*
   * Transfer Structure factor F_g [A]
   *  to potential V_g  [V] ( = F_g / (GE * bv.vc) )
   *
   * V_g does not include relativistic correction Gamma.
   */
  return c_div( fg, c_conv(GE * bv.vc, 0.0) );
}


complex vg_to_fg(complex vg, CELL_PAR bv)
{
  /*
   * Transfer potential V_g [V]  ( = F_g / (GE * bv.vc) )
   *  to Structure factor F_g
   *
   * V_g must not include relativistic correction Gamma.
   */
  return c_mul( vg, c_conv(GE * bv.vc, 0.0) );
}


complex fg_to_ug(complex fg, double accel_kvol, CELL_PAR bv) 
{
  /*
   * Transfer Structure factor F_g
   *  to potential U_g  [A^{-2}] ( = F_g / (PI * bv.vc) )
   *
   * U_g includes relativistic correction Gamma.
   */
  return c_div( fg, c_conv( PI * bv.vc / relativ_corr(accel_kvol), 0.0) );
}


complex ug_to_fg(complex ug, double accel_kvol, CELL_PAR bv) 
{
  /*
   * Transfer potential U_g  [A^{-2}] ( = F_g / (PI * bv.vc) )
   *  to Structure factor F_g
   *
   * U_g includes relativistic correction Gamma.
   */
  return c_mul( ug, c_conv( PI * bv.vc / relativ_corr(accel_kvol), 0.0) );
}


double fg_to_xig(complex fg, double accel_kvol, CELL_PAR bv)
{
  /*
   * Transfer Structure factor F_g
   *  to Absolute value of xi_g [A]
   */
  double t;
  
  t = c_abs( fg_to_ug( fg, accel_kvol, bv) );
  if ( t < 1e-12 ){
    return XIGMAX;	/* forbidden */
  }
  else{
    return 1.0 / ( wavelength(accel_kvol) * t );
  }
}


double fg_to_xig2(
  complex fg, double relativistic_correction, double wave_length, CELL_PAR bv
  )
{
  /*
   * Transfer Structure factor F_g
   *  to Absolute value of xi_g [A]
   */
  double t;
  
  t = c_abs( c_div( fg, c_conv( PI * bv.vc / relativistic_correction, 0.0) ) );
  if ( t < 1e-12 ){
    return XIGMAX;	/* forbidden : artificial */
  }
  else{
    return 1.0 / ( wave_length * t );
  }
}


double vg_to_xig( complex vg, double relativistic_correction, double wave_length )
{
  /*
   * Transfer Vg to absolute value of xi_g [A]
   */
  double t;
  
  t = c_abs( c_div( vg, c_conv( PI / GE / relativistic_correction, 0.0) ) );
  if ( t < 1e-12 ){
    return 1e+10;	/* forbidden : artificial */
  }
  else{
    return 1.0 / ( wave_length * t );
  }
}


complex vg_to_ug( complex vg, double relativistic_correction )
{
  /*
   * Transfer Vg [V] to Ug [A^{-2}].
   */
  return c_mul( vg, c_conv( relativistic_correction * GE / PI, 0.0 ) );
}


complex vg_to_ug_2( complex vg, double accel_kvol )
{
  /*
   * Transfer Vg [V] to Ug [A^{-2}].
   */
  return c_mul( vg, c_conv( 0.006648352 * (1.0 + 1.956934e-3 * accel_kvol), 0.0 ) );
}


complex ug_to_vg( complex ug, double relativistic_correction )
{
  /*
   * Transfer Ug [A^{-2}] to Vg [V].
   */
  return c_mul( ug, c_conv( PI / (relativistic_correction * GE), 0.0 ) );
}


/* --------------------------------------------------------------------------- */


SYM_OP decode_sym_op(char *str)
{
  /*
   * Decode symmetry operations of the spgrp.dat
   *   into rotation matrix and translation vector
   */
  SYM_OP rt;
  char *delims=",";
  char *token, *p, s1[2], s2[2];
  char str_x[10], str_y[10], str_z[10];
  
  /* Divide str into x, y and z parts */
  token = strtok(str,delims);		strcpy(str_x,token);	/* x part */
  token = strtok(NULL,delims);	strcpy(str_y,token);	/* y part */
  token = strtok(NULL,delims);	strcpy(str_z,token);	/* z part */
  
  /* Initialize */
  rt.r11 = 0.0; rt.r12 = 0.0; rt.r13 = 0.0;
  rt.r21 = 0.0; rt.r22 = 0.0; rt.r23 = 0.0;
  rt.r31 = 0.0; rt.r32 = 0.0; rt.r33 = 0.0;
  rt.t1  = 0.0; rt.t2  = 0.0; rt.t3  = 0.0;
  
  /* Rotation matrix */
  /* X  */
  if( strchr(str_x,'X') != NULL ){
    if( (p = strchr(str_x,'-')) != NULL && *(++p) == 'X'){
      rt.r11 = -1.0;
    }
    else{
      rt.r11 = 1.0;
    }	
  }
  if( strchr(str_x,'Y') != NULL ){
    if( (p = strchr(str_x,'-')) != NULL && *(++p) == 'Y'){
      rt.r12 = -1.0;
    }
    else{
      rt.r12 = 1.0;
    }	
  }
  if( (p = strchr(str_x,'Z')) != NULL ){
    if( (p = strchr(str_x,'-')) != NULL && *(++p) == 'Z'){
      rt.r13 = -1.0;
    }
    else{
      rt.r13 = 1.0;
    }	
  }
  
  /* Y  */
  if( strchr(str_y,'X') != NULL ){
    if( (p = strchr(str_y,'-')) != NULL && *(++p) == 'X'){
      rt.r21 = -1.0;
    }
    else{
      rt.r21 = 1.0;
    }
  }
  if( strchr(str_y,'Y') != NULL ){
    if( (p = strchr(str_y,'-')) != NULL && *(++p) == 'Y'){
      rt.r22 = -1.0;
    }
    else{
      rt.r22 = 1.0;
    }	
  }
  if( (p = strchr(str_y,'Z')) != NULL ){
    if( (p = strchr(str_y,'-')) != NULL && *(++p) == 'Z'){
      rt.r23 = -1.0;
    }
    else{
      rt.r23 = 1.0;
    }	
  }
  
  /* Z  */
  if( strchr(str_z,'X') != NULL ){
    if( (p = strchr(str_z,'-')) != NULL && *(++p) == 'X'){
      rt.r31 = -1.0;
    }
    else{
      rt.r31 = 1.0;
    }	
  }
  if( strchr(str_z,'Y') != NULL ){
    if( (p = strchr(str_z,'-')) != NULL && *(++p) == 'Y'){
      rt.r32 = -1.0;
    }
    else{
      rt.r32 = 1.0;
    }	
  }
  if( (p = strchr(str_z,'Z')) != NULL ){
    if( (p = strchr(str_z,'-')) != NULL && *(++p) == 'Z'){
      rt.r33 = -1.0;
    }
    else{
      rt.r33 = 1.0;
    }	
  }
  
  /* Translation */
  if ( (p = strchr(str_x,'/')) != NULL){
    s1[0] = *(p-1); s1[1] = '\0'; s2[0] = *(p+1); s2[1] = '\0';
    rt.t1 = atof( s1 ) / atof( s2 );
  }
  if ( (p = strchr(str_y,'/')) != NULL){
    s1[0] = *(p-1); s1[1] = '\0'; s2[0] = *(p+1); s2[1] = '\0';
    rt.t2 = atof( s1 ) / atof( s2 );
  }
  if ( (p = strchr(str_z,'/')) != NULL){
    s1[0] = *(p-1); s1[1] = '\0'; s2[0] = *(p+1); s2[1] = '\0';
    rt.t3 = atof( s1 ) / atof( s2 );
  }
  return rt;
}


SYM_OP inv_sym_op( SYM_OP r )
{
  SYM_OP ir;
  
  ir.r11 = - r.r11;	ir.r12 = - r.r12;	ir.r13 = - r.r13;
  ir.r21 = - r.r21;	ir.r22 = - r.r22;	ir.r23 = - r.r23;
  ir.r31 = - r.r31;	ir.r32 = - r.r32;	ir.r33 = - r.r33;
  ir.t1 = - r.t1;		ir.t2 = - r.t2;		ir.t3 = - r.t3;
  
  return ir;
}


SYM_OP *read_sp_group_tbl(
  int num_sp_group, int *num_sym_op, int *inversion, char *lattice_type
  )
{
  /*
   * symmetry oparation
   *  space group number : target
   */
  int j, n_spgrp, laue_group, n_sym;
  char space_group[10], buf[128], *delims=" ,\t", *token;
  SYM_OP *rt;
  FILE *fp;
  char mbfit_space_group_file[256];
  sprintf(mbfit_space_group_file,"%s/mbfit/mbfitspg.dat",PREFIX);

    
  /* printf("Read space-group table.\n"); */
  if( (fp = fopen(mbfit_space_group_file,"r")) == NULL ){
    printf("Cannot open %s.\n", SPACE_G_FILE);
    exit(EXIT_FAILURE);
  }
  
  while( comment_skip_fgets(100, buf, fp) != NULL ){ /* 1st line of each space group */
    token = strtok(buf, delims);	n_sym = atoi(token);
    token = strtok(NULL,delims);	*inversion = atoi(token);
    token = strtok(NULL,delims);	n_spgrp = atoi(token);
    
    if(n_spgrp == num_sp_group){		/* Found */
      token = strtok(NULL,delims);	laue_group = atoi(token);
      token = strtok(NULL,delims);	strcpy(space_group, token);
      lattice_type[0] = space_group[0];	lattice_type[1] = '\0';
      
      rt = new_vector( SYM_OP, n_sym * 8 );	chk_vector(rt);
      for(j=0;j<n_sym;j++){
        comment_skip_fgets( 100, buf,fp);
        rt[j] = decode_sym_op(buf);
      }
      
      /* lattice_type  */
      if (space_group[0]=='P'){
      }
      else if (  space_group[0]=='I'){
        for(j=0;j<n_sym;j++){
          rt[j+n_sym] = rt[j];
          rt[j+n_sym].t1 = rt[j].t1 + 0.5;
          rt[j+n_sym].t2 = rt[j].t2 + 0.5;
          rt[j+n_sym].t3 = rt[j].t3 + 0.5;
        }
        n_sym *= 2;
      }
      else if ( space_group[0]=='A' ){
        for(j=0;j<n_sym;j++){
          rt[j+n_sym] = rt[j];
          rt[j+n_sym].t2 = rt[j].t2 + 0.5;
          rt[j+n_sym].t3 = rt[j].t3 + 0.5;
        }
        n_sym *= 2;
      }
      else if ( space_group[0]=='B' ){
        for(j=0;j<n_sym;j++){
          rt[j+n_sym] = rt[j];
          rt[j+n_sym].t1 = rt[j].t1 + 0.5;
          rt[j+n_sym].t3 = rt[j].t3 + 0.5;
        }
        n_sym *= 2;
      }
      else if ( space_group[0]=='C' ){
        for(j=0;j<n_sym;j++){
          rt[j+n_sym] = rt[j];
          rt[j+n_sym].t1 = rt[j].t1 + 0.5;
          rt[j+n_sym].t2 = rt[j].t2 + 0.5;
        }
        n_sym *= 2;
      }
      else if (space_group[0]=='R'){
        /* for(j=0;j<n_sym;j++){
					rt[j+n_sym] = rt[j];
					rt[j+n_sym].t1 = rt[j].t1 + 0.6666667;
					rt[j+n_sym].t2 = rt[j].t2 + 0.3333333;
					rt[j+n_sym].t3 = rt[j].t3 + 0.3333333;
				}
				for(j=0;j<n_sym;j++){
					rt[j+n_sym*2] = rt[j];
					rt[j+n_sym*2].t1 = rt[j].t1 + 0.3333333;
					rt[j+n_sym*2].t2 = rt[j].t2 + 0.6666667;
					rt[j+n_sym*2].t3 = rt[j].t3 + 0.6666667;
				}
				n_sym *= 3;*/
      }
      else if(space_group[0]=='F'){
        for(j=0;j<n_sym;j++){
          rt[j+n_sym] = rt[j];
          rt[j+n_sym].t2 = rt[j].t2 + 0.5;
          rt[j+n_sym].t3 = rt[j].t3 + 0.5;
        }
        for(j=0;j<n_sym;j++){
          rt[j+n_sym*2] = rt[j];
          rt[j+n_sym*2].t1 = rt[j].t1 + 0.5;
          rt[j+n_sym*2].t3 = rt[j].t3 + 0.5;
        }
        for(j=0;j<n_sym;j++){
          rt[j+n_sym*3] = rt[j];
          rt[j+n_sym*3].t1 = rt[j].t1 + 0.5;
          rt[j+n_sym*3].t2 = rt[j].t2 + 0.5;
        }
        n_sym *= 4;
      }
      else{
        fprintf(stderr,"Invalid lattice type in space-group table.\n");
        exit(EXIT_FAILURE);
      }
      
      if( *inversion == 1){			/* Centrosymmetric */
        for(j=n_sym; j<n_sym * 2; j++){
          rt[j] = inv_sym_op( rt[j-n_sym] );
        }
        n_sym *= 2;
      }
      
      *num_sym_op = n_sym;
      return rt;
    }
    else{
      for(j=0;j<n_sym;j++){
        comment_skip_fgets( 100, buf, fp);	/* skip */
      }
    }
  }
  
  printf(
    "$$ Space group no. = %d is not found in the table.\n",
    num_sp_group
    );
  exit(EXIT_FAILURE);
  fclose(fp);
  return rt;	/* <<<<< dummy */
}


void read_latt_type( int num_sp_group, char *lattice_type)
{
  int n_spgrp;
  char space_group[10];
  char buf[128];
  char *delims=" ,\t";
  char *token;
  FILE *fp;
  
  /* printf("Read space-group table (for lattice_type).\n"); */
   char mbfit_space_group_file[256];
  sprintf(mbfit_space_group_file,"%s/source/mbfitspg.dat",PREFIX);
 
  if( (fp = fopen(mbfit_space_group_file,"r")) == NULL ){
    printf("Cannot open %s.\n",SPACE_G_FILE);
    exit(EXIT_FAILURE);
  }
  
  while( comment_skip_fgets(100, buf, fp) != NULL ){ /* 1st line of each space group */
    token = strtok(buf, delims);	/*	n_sym = atoi(token);*/
    token = strtok(NULL,delims);	/* *inversion = atoi(token);*/
    token = strtok(NULL,delims);	n_spgrp = atoi(token);
    if(n_spgrp == num_sp_group){		/* Found */
      token = strtok(NULL,delims);	/* laue_group = atoi(token); */
      token = strtok(NULL,delims);	strcpy(space_group, token);
      fclose(fp);
      space_group[1] = '\0';
      strcpy(lattice_type, space_group);
      return;
    }
  }
  printf("Not found.\n");
  fclose(fp);	
}


HKL gen_reflex( HKL g, SYM_OP rt )
     /*
      * Generate new reflections by symmetry operation
      */
{
  HKL gg;
  
  gg.h = g.h * (int) rt.r11 + g.k * (int) rt.r21 + g.l * (int) rt.r31;
  gg.k = g.h * (int) rt.r12 + g.k * (int) rt.r22 + g.l * (int) rt.r32;
  gg.l = g.h * (int) rt.r13 + g.k * (int) rt.r23 + g.l * (int) rt.r33;
  return gg;
}


XYZ gen_pos( XYZ t, SYM_OP rt )
     /*
      * Generate new atom positions by symmetry operation
      */
{
  XYZ k;
  
  k.x = t.x * rt.r11 + t.y * rt.r12 + t.z * rt.r13 + rt.t1;
  k.y = t.x * rt.r21 + t.y * rt.r22 + t.z * rt.r23 + rt.t2;
  k.z = t.x * rt.r31 + t.y * rt.r32 + t.z * rt.r33 + rt.t3;
  return k;
}


double fractional( double x )
     /*
      * fractional coordinate t: (0 <= t < 1.0)
      */
{
  double t;
  
  t = x - floor(x);
  while ( t < 0.0 ||  t >= 1.0){
    if ( t < 0.0 ){ t += 1.0; }
    else if (t >= 1.0) { t -= 1.0; }
  }
  return t;
}


int judge_two_atom(XYZ t, XYZ k)
{
  double dx, dy, dz;
  
  /* printf("t=%lf %lf %lf,  k=%lf %lf %lf\n",t.x,t.y,t.z, k.x,k.y,k.z); */
  dx = fabs(  fractional(t.x) -  fractional(k.x) );
  dy = fabs(  fractional(t.y) -  fractional(k.y) );
  dz = fabs(  fractional(t.z) -  fractional(k.z) );
  /* printf("  dx,dy,dz = %lf %lf %lf\n",dx,dy,dz); */
  
  if ( ( dx < POSEPS || fabs(dx - 1.0) < POSEPS )
       && ( dy < POSEPS || fabs(dy - 1.0) < POSEPS )
       && ( dz < POSEPS || fabs(dz - 1.0) < POSEPS ) ){
    return 1;			/* The two atoms have approximately the same position. */
  }
  else{
    return 0;
  }
}


int special_pos_check(int n, XYZ *pos)
     /*
      * Special position check.
      *   (omit duplicate positions)
      */
{
  XYZ *p;
  int i, j, ct;
  
  p = new_vector(XYZ, n);		chk_vector(p);
  
  ct = 0;
  for(i=0;i<n;i++){
    for(j=0;j<ct;j++){
      if( judge_two_atom(pos[i],p[j]) == 1 ){	/* same coord. */
        goto NEXT_POS;
      }
    }
    p[ct++] = pos[i];
  NEXT_POS: continue;
  }
  for(i=0;i<ct;i++){
    pos[i] = p[i];	/* pos[] were reduced to special positions p[] */
  }
  
  free_vector(p);
  return ct;		/* number of sites */
}


void atom_multiplicity(
  int n_atom, ATOM *atm, int num_sym_op, SYM_OP *sym_op
  )
     /*
      * Evaluation of multiplicity of atom positions
      */
{
  int i, j, ct_pos, num_indep_pos;
  XYZ *tpos, pos;
  
  tpos = new_vector(XYZ, num_sym_op * 2);	chk_vector(tpos);
  for(i=0;i<n_atom;i++){
    ct_pos = 0;
    pos.x = atm[i].x; pos.y = atm[i].y; pos.z = atm[i].z; /* sym = atm[i].sym; */
    /* pos = gen_pos(pos, sym); */
    for(j=0; j<num_sym_op; j++){
      tpos[ct_pos++] = gen_pos(pos, sym_op[j]);
    }
    num_indep_pos = special_pos_check(ct_pos, tpos);
    
    if(epsilon.auto_multiplicity_off == 0){
      atm[i].mul = (double) num_indep_pos / (double) num_sym_op;
    }
    else if(epsilon.auto_multiplicity_off == 1){
      atm[i].mul = 1.0;
    }
    else{
      printf(
        "$$ invalid auto_multiplicity_off=%d\n. Abort.\n",
        epsilon.auto_multiplicity_off
        );
      exit(EXIT_FAILURE);
    }
  }
  free_vector(tpos);
}


/* ----- anisotropic B ----- */

int ij_to_i(int ij)
{
  if      (ij == 11){ return 0; }
  else if (ij == 22){ return 1; }
  else if (ij == 33){ return 2; }
  else if (ij == 12){ return 3; }
  else if (ij == 23){ return 4; }
  else if (ij == 13){ return 5; }
  else{
    printf("$$ Invalid ij. Abort.\n");
    exit(EXIT_FAILURE);
    return 0;
  }
}

#if 0
double mul_aniso_bij(int ij, ATOM *atm)
{
  int iv;
  
  iv = ij_to_i(ij);
  return    atm->bc[iv][ij_to_i(11)] * atm->b11
    + atm->bc[iv][ij_to_i(22)] * atm->b22
      + atm->bc[iv][ij_to_i(33)] * atm->b33
        + atm->bc[iv][ij_to_i(12)] * atm->b12
          + atm->bc[iv][ij_to_i(23)] * atm->b23
            + atm->bc[iv][ij_to_i(13)] * atm->b13;
}
#endif

#if 0
void cst_aniso_bij(ATOM *atm)
{
  if(atm->bf == 1){
    atm->b11 = mul_aniso_bij(11, atm);
    atm->b22 = mul_aniso_bij(22, atm);
    atm->b33 = mul_aniso_bij(33, atm);
    atm->b12 = mul_aniso_bij(12, atm);
    atm->b23 = mul_aniso_bij(23, atm);
    atm->b13 = mul_aniso_bij(13, atm);
  }
}
#endif

double bij_to_beq(CELL_PAR bv, ATOM *atm)
     /*
      * bij to equivalent b
      *
      *   B_equiv = Tr( beta_ij * real_metric_tensor ) * 4 / 3
      *   (C. Giacovazzo et al: "Fundamentals of Crystallography", 1992, IUCR, Oxford Univ. Press)
      */
{
  if(atm->bf == 1){
    return ( atm->b11 * bv.aa + atm->b22 * bv.bb + atm->b33 * bv.cc
             + 2 * atm->b12 * bv.ab + 2 * atm->b23 * bv.bc +2 * atm->b13 * bv.ac ) * 4.0 / 3.0;
  }
  else{
    return atm->b;
  }
}


void bij_to_b(HKL g, double sq_s, ATOM *atm, CELL_PAR bv)
     /*
      * bij to b
      *  exp(-b*sq_s) = exp(-b11*h-b22*k-b33*l-2*b12*h*k-2*b23*k*l-2*b13*h*l)
      *
      */
{
  double b;
  
  if(atm->bf == 1){
    if ( sq_s > epsilon.eps ){
      b =	(
        atm->b11 * sq((double)g.h)
        + atm->b22 * sq((double)g.k)
        + atm->b33 * sq((double)g.l)
        + atm->b12 * 2.0 * (double)g.h * (double)g.k
        + atm->b23 * 2.0 * (double)g.k * (double)g.l
        + atm->b13 * 2.0 * (double)g.l * (double)g.h
        ) / sq_s;
      /*
      printf("$$\n");
      printf("$$ g = %d %d %d\n", g.h, g.k, g.l);
      printf("$$ atm.bf = %d\n", atm->bf);
      printf("$$ sq_s = %lf \n", sq_s);
      printf("$$ b11,22,33,12,23,13: %lf %lf %lf %lf %lf %lf\n",
             atm->b11, atm->b22, atm->b33, atm->b12, atm->b23, atm->b13);
       */
    }
    else {
      b = bij_to_beq(bv, atm);	/* average */
    }
    atm->b = b;
    /*	printf("$$ converted b = %lf\n", b); */
  }
}


/* --------------------------------------------------------------------------- */


int search_g_table(int n_gtable, HKL g, HKL *gtable )
     /*
      * Search elements of g from g_table.
      * If g is not found in g_table, return -1.
      */
{
  int i;
  
  for(i=0; i<n_gtable; i++){
    if( g.h == gtable[i].h && g.k == gtable[i].k && g.l == gtable[i].l ){
      return i;		/* found */
    }
  }
  return -1;		/* not found */
}


void stock_fg_tbl_sym(
  HKL g, complex fg,
  int *ct_fg_table, int g_table_size, HKL *g_table, complex *fg_table,
  int num_sym_op, SYM_OP *sym_op
  )
     /*
      * Stock fg into fg_table with those of symmetrically equivalent g
      */
{
  int i;
  HKL  rf;
  double rad;
  complex phase;
  
  if ( *ct_fg_table < g_table_size ){
    g_table[*ct_fg_table].h = g.h;
    g_table[*ct_fg_table].k = g.k;
    g_table[*ct_fg_table].l = g.l;
    fg_table[*ct_fg_table] = fg;
    (*ct_fg_table)++;
  }
  else{
    return;
  }
  
  for(i=0;i<num_sym_op;i++){		/* Symmetry operations of reflections */
    if ( *ct_fg_table < g_table_size ){
      rf = gen_reflex( g, sym_op[i] );
      /*	printf("$$ (stock_fg_tbl_sym) g(%d) = %d %d %d\n", i, rf.h, rf.k, rf.l);*/
      if( search_g_table( *ct_fg_table, rf, g_table) < 0 ){ 	/* Not found */
        g_table[*ct_fg_table].h = rf.h;
        g_table[*ct_fg_table].k = rf.k;
        g_table[*ct_fg_table].l = rf.l;
        rad = 2.0 * PI * (
          (double) g.h * sym_op[i].t1 +
          (double) g.k * sym_op[i].t2 +
          (double) g.l * sym_op[i].t3 
          );
        phase = c_exp( c_conv(0.0, rad) );
        fg_table[ (*ct_fg_table)++ ] = c_mul( fg, phase );
        /*printf("g=%d %d %d phase=%lf %lf\n",rf.h,rf.k,rf.l,phase.re, phase.im);*/
      }
    }
    else{
      return;
    }
  }
  return;
}


/* --------------------------------------------------------------------------- */


complex strfct_xdyl(
  int n_atom, int num_nthatm[],
  double b_factor[], double occupancy[], double f_absrp_0[],
  HKL g, XYZ **pos, XDT *xdyl, CELL_PAR bv
  )
     /*
      * Structure factor F_g [A]
      */
{
  int i,j;
  double sq_s, phase, f_elastic, f_absorption;
  complex fg, gsf, ztmp;
  R_HKL rg;
  
  rg.h = (double) g.h; rg.k = (double) g.k; rg.l = (double) g.l;
  sq_s = sq_abs(rg, bv) / 4.0;	/*  s = sin(theta) / lambda = 1 / (2d) = g / 2  */
  
  fg = c_conv(0.0,0.0);
  for(i = 0; i < n_atom; i++){
    f_elastic = atomsf_xdt( sq_s, xdyl[i] );
    f_absorption = atom_f_absorption( sq_s, f_absrp_0[i], b_factor[i] );
    gsf = c_conv(0.0,0.0);
    for(j = 0; j < num_nthatm[i]; j++){
      phase = - 2.0 * PI * (
        rg.h * pos[i][j].x + rg.k * pos[i][j].y + rg.l * pos[i][j].z );
      gsf = c_add( gsf, c_exp( c_conv(0.0, phase) ) );
    }
    ztmp = c_mul( c_conv( occupancy[i] * exp( - b_factor[i] * sq_s), 0 ),
                  c_conv( f_elastic, f_absorption ) );
    fg = c_add( fg, c_mul( gsf, ztmp ) );
  }
  return fg;
}


complex strfct_ext(
  int n_atom, int num_nthatm[],
  double b_factor[], double occupancy[], double f_absrp_0[],
  HKL g, XYZ **pos, XDT *xdyl, CELL_PAR bv, char *lattice_type
  )
     /*
      * Calc. of structure factor with extinction rule
      */
{
  complex fg;
  
  if( extinction_rule(g, lattice_type) == 0 ){
    fg = strfct_xdyl(
      n_atom, num_nthatm, b_factor, occupancy, f_absrp_0, g,  pos, xdyl, bv
      );
  }
  else{
    fg = c_conv(0,0);
  }
  return fg;
}


void newpos_sym_op(
  int n_atom, ATOM *atm, int *num_nthatm, XYZ **pos, int num_sym_op, SYM_OP *sym_op
  )
     /*
      * generate atom positions by symmetry operations
      * (with checking special positions)
      */
{
  int i,j,ct_pos;
  XYZ *tpos, t;
  
  /* printf( "num_sym_op=%d\n", num_sym_op );*/
  tpos = new_vector( XYZ, num_sym_op * 2 );	chk_vector(tpos);
  for(i=0;i<n_atom;i++){
    ct_pos = 0;
    t.x = atm[i].x; t.y = atm[i].y; t.z = atm[i].z; /* sym = atm[i].sym; */
    /* t = gen_pos(t, sym);	*/						/* <<<<< */
    for(j=0; j<num_sym_op; j++){
      tpos[ct_pos++] = gen_pos(t, sym_op[j]);
    }
    
    if(epsilon.auto_multiplicity_off == 0){
      num_nthatm[i] = special_pos_check(ct_pos, tpos);
      /* printf("num_nthatm[%d]=%d\n",i,num_nthatm[i]); */
      for(j=0; j< num_nthatm[i]; j++){
        pos[i][j] = tpos[j];
        /*printf("pos[%d][%d]=%lf %lf %lf\n",i,j,tpos[j].x,tpos[j].y,tpos[j].z);*/
      }
    }
    else if(epsilon.auto_multiplicity_off == 1){
      num_nthatm[i] = num_sym_op;
      for(j=0; j< num_nthatm[i]; j++){
        pos[i][j] = tpos[j];
        /*printf("pos[%d][%d]=%lf %lf %lf\n",i,j,tpos[j].x,tpos[j].y,tpos[j].z);*/
      }
    }
    else{
      printf("$$ Invalid epsilon.auto_multiplicity_off. Abort.\n");
      exit(EXIT_FAILURE);
    }
  }
  free_vector(tpos);
}


complex lofg_to_fg_0(int lof_type, LO_FG fg_lo)
     /*
      * Low order structure factor
      */
{
  complex fg;
  
  if(lof_type == 1){
    fg.re = fg_lo.re;
    fg.im = fg_lo.im;
  }
  else if(lof_type == 2){
    printf("$$ Sorry. lof_type=2 is not supported in lofg_to_fg_0. Abort.\n");
    exit(EXIT_FAILURE);
    /* */
  }
  else if(lof_type == 3  || lof_type == 4 || lof_type == 5){
    printf("$$ Sorry. lof_type=3, 4 and 5 are not supported in lofg_to_fg_0. Abort.\n");
    exit(EXIT_FAILURE);
  }
  else{
    printf("$$ Invalid lof_type. Abort.\n");	exit(EXIT_FAILURE);
  }
  return fg;
}


#if 0
complex str_fct(
  int ind, HKL g,
  char *lattice_type, int num_sym_op, SYM_OP *sym_op,
  int n_atom, ATOM *atm, XDT *xdyl,
  int num_low_order_g, HKL *g_low_order, int lof_type, LO_FG *fg_low_order,
  int *ct_fg_table, HKL *g_table, complex *fg_table, int maxnum_fg_table,
  CELL_PAR bv
  )
{
  /*
   * Structure factor
   *   with absorption and without relativistic correction
   * Table-management system
   *
   * ind < 0: initialize (generate atom positions)
   * ind > 0: calculation
   * ind = 0: quit (free matrix)
   *
   * fg_low_order must not include relativistic correction.
   */
  int i, ng;
  complex fg;
  static int *num_nthatm;
  static double *b_factor, *occupancy, *f_absrp_0;
  static XYZ **pos;

  if (ind < 0) {/* Initialize */
    num_nthatm	= new_vector(int, n_atom);  chk_vector(num_nthatm);
    pos 		= new_matrix(XYZ, n_atom, num_sym_op);		chk_matrix(pos);
    b_factor	= new_vector(double, n_atom);				chk_vector(b_factor);
    occupancy	= new_vector(double, n_atom);				chk_vector(occupancy);
    f_absrp_0	= new_vector(double, n_atom);				chk_vector(f_absrp_0);
    
    newpos_sym_op(n_atom, atm, num_nthatm, pos, num_sym_op, sym_op);
    for(i=0; i<n_atom; i++){
      if(atm[i].bf == 0) {
        b_factor[i] = atm[i].b;							/* Isotropic B-factor */
      }
      else{												/* Anisotropic B-factor */
        fprintf(stderr,"Anisotropic B-factor: sorry, now preparing.\n");
        exit(EXIT_FAILURE);
      }
      occupancy[i] = atm[i].occ;							/* Occupation */
      f_absrp_0[i] = atm[i].fabs;							/* Absorption coef. (s=0) */
    }
    *ct_fg_table = 0;
    for(i=0; i<num_low_order_g; i++){
      stock_fg_tbl_sym(
        g_low_order[i], lofg_to_fg_0(lof_type, fg_low_order[i]), ct_fg_table,
        maxnum_fg_table, g_table, fg_table, num_sym_op, sym_op
        );
    }
    return;
  }
  else if (ind == 0){					/* Free */
    free_vector(num_nthatm);
    free_matrix(pos);
    free_vector(b_factor);
    free_vector(occupancy);
    free_vector(f_absrp_0);
    return;
  }
  
  /*
   * Calc. or search of structure factors (ind > 0)
   */
  if ( ( ng = search_g_table( *ct_fg_table, g, g_table) ) >= 0 ){	/* found */
    fg = fg_table[ng];
  }
  else{														/* not found in table */
    fg = structure_factor_ext( n_atom, num_nthatm, b_factor, occupancy, f_absrp_0,
                               g,  pos, xdyl, bv, lattice_type );
    stock_fg_tbl_sym( g, fg, ct_fg_table, maxnum_fg_table,
                      g_table, fg_table, num_sym_op, sym_op );
  }
  return fg;
}
#endif


/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
complex calc_structure_factor_ph(
  HKL g, double sq_s, int num_sym_op, SYM_OP sym_op[],
  int n_atom, ATOM atm[], double f_elastic[], double f_inelastic[]
  )
{ 
  /*
   * Calc. of structure factor F_g [A] adding phases
   *   using symmetry operation of reflections
   *   and given f_elastic[] and f_inelastic[]
   * (s = sin(theta) / lambda = 1 / (2d) = g / 2)
   */
 int i, j;
  double	phase, temp;
  complex fg, gsf, ztmp;
  HKL sym_g;
  XYZ pos;
  
  fg = c_conv(0.0,0.0);
  for(i=0; i<n_atom; i++){
    gsf = c_conv(0.0,0.0);
    /* Constraint of coordinates (special position) */
    pos.x = atm[i].x;  pos.y = atm[i].y;  pos.z = atm[i].z;
    /* sym = atm[i].sym; */
    /* pos = gen_pos(pos, sym); */
    if (atm[i].bf == 0){  /* Isotropic B */
      for(j=0; j<num_sym_op; j++){
        sym_g = gen_reflex(g, sym_op[j]);
        phase = - 2.0 * PI *
          (
            (double) sym_g.h * pos.x
            + (double) sym_g.k * pos.y
            + (double) sym_g.l * pos.z
            + (double) g.h * sym_op[j].t1
            + (double) g.k * sym_op[j].t2
            + (double) g.l * sym_op[j].t3
            );
        ztmp = c_exp(c_conv(0.0, phase));
        gsf = c_add(gsf, ztmp);
      }
      temp = exp(- atm[i].b * sq_s);
      ztmp = c_mul(
        c_conv( atm[i].occ * temp * atm[i].mul, 0.0 ),
        c_conv( f_elastic[i], f_inelastic[i] )
        );
      fg = c_add(fg, c_mul(gsf, ztmp));
    }
    else if (atm[i].bf == 1){  /* Anisotropic B */
      for(j=0; j<num_sym_op; j++){
        sym_g = gen_reflex(g, sym_op[j]);
        phase = - 2.0 * PI *
          (
            (double) sym_g.h * pos.x
            + (double) sym_g.k * pos.y
            + (double) sym_g.l * pos.z
            + (double) g.h * sym_op[j].t1
            + (double) g.k * sym_op[j].t2
            + (double) g.l * sym_op[j].t3
            );
        temp = exp(
          - (  atm[i].b11 * sq((double)sym_g.h)
               + atm[i].b22 * sq((double)sym_g.k)
               + atm[i].b33 * sq((double)sym_g.l)
               + atm[i].b12 * 2.0 * (double)sym_g.h * (double)sym_g.k
               + atm[i].b23 * 2.0 * (double)sym_g.k * (double)sym_g.l
               + atm[i].b13 * 2.0 * (double)sym_g.l * (double)sym_g.h )
          );
        ztmp = c_mul(c_exp(c_conv(0.0, phase)), c_conv(temp, 0));
        gsf = c_add(gsf, ztmp);
      }
      ztmp = c_mul(
        c_conv( atm[i].occ * atm[i].mul, 0.0 ),
        c_conv( f_elastic[i], f_inelastic[i] )
        );
      /* printf(
          "$$ %d %d %d: Reduced iso_b = %lf\n", g.h, g.k, g.l, atm[i].b
         ); */
      fg = c_add(fg, c_mul(gsf, ztmp));
    }
    else{
      printf("ERROR: Invalid atom.bf. Abort.\n");
      exit(EXIT_FAILURE);
    }
  }
  return fg;
}


complex calc_structure_factor(
  HKL g, int num_sym_op, SYM_OP sym_op[],
  int n_atom, ATOM atm[], XDT xdyl[], ATOMF_BK ftbl[],
  double kvolts, CELL_PAR bv
  )
{
  int i;
  double *f_ela, *f_inela, sq_s, s;
  complex Fg, atomf;
  
  /*    s = sin(theta) / lambda = 1 / (2d) = g / 2   */
  sq_s = sq_abs( hkl2rhkl(g), bv) / 4.0;
  s = sqrt(sq_s);
  
  f_ela = new_vector(double, n_atom);    chk_vector(f_ela);
  f_inela = new_vector(double, n_atom);  chk_vector(f_inela);
  
  for(i=0; i<n_atom; i++){
    if(atm[i].af == 0){ /* Mott formula + BK's absorption */
      bij_to_b( g, sq_s, &(atm[i]), bv );
      atom_bk( ftbl[i], s, atm[i].b, &atomf, kvolts );
      f_ela[i] = atomsf_xdt( sq_s, xdyl[i] );
      f_inela[i] = atomf.im;
      /* printf(
        "$$s=%lf, asf_bk=%lf, asf_xdt=%lf\n",
        sqrt(sq_s), atomf.re, f_ela[i]
        );*/
    }
    else if(atm[i].af == 1){ /* BK: Bird & King */
      bij_to_b( g, sq_s, &(atm[i]), bv );
      atom_bk( ftbl[i], s, atm[i].b, &atomf, kvolts );
      f_ela[i] = atomf.re;
      f_inela[i] = atomf.im;
    }
    else{
      printf("$$ Invalid af. Abort.\n");	exit(EXIT_FAILURE);
    }
  }
  
  Fg = calc_structure_factor_ph( 
    g, sq_s, num_sym_op, sym_op, n_atom, atm, f_ela, f_inela
    );
  free_vector( f_ela );
  free_vector( f_inela );
  return Fg;
}


double calc_F_0_re(
  int num_sym_op, int n_atom, ATOM atm[], XDT xdyl[], ATOMF_BK ftbl[],
  CELL_PAR bv
  )
{
  /* F_0.re (for V_0 and U_0) */
  int i;
  double F0_re, f_ela, b=0.0, kvolts=100.0;
  complex atomf;
  
  F0_re = 0.0;
  for(i=0; i<n_atom; i++){
    if(atm[i].af == 0){	/* Mott formula */
      f_ela	= atomsf_xdt( 0.0, xdyl[i] );
    }
    else if(atm[i].af == 1 || atm[i].af == 2){  /* Bird & King */
      atom_bk( ftbl[i], 0.0, b, &atomf, kvolts );
      /* b & kvolts are dummy args. */
      f_ela	= atomf.re;
    }
    else{
      printf("$$ Invalid af. Abort.\n");	exit(EXIT_FAILURE);
    }
    F0_re += f_ela * atm[i].occ * num_sym_op * atm[i].mul;
  }
  return F0_re;
}


complex structure_factor_ext(
  HKL g, int num_sym_op, SYM_OP sym_op[],
  int n_atom, ATOM atm[], XDT xdyl[], ATOMF_BK ftbl[],
  double kvolts, CELL_PAR bv, char *lattice_type
  )
{
  /*
   * Calc. of structure factor with extinction rule
   */
  complex fg;
  
  if( extinction_rule(g, lattice_type) == 0 ){
    fg = calc_structure_factor(
      g, num_sym_op, sym_op, n_atom, atm, xdyl, ftbl, kvolts, bv
      );
  }
  else{
    fg = c_conv(0.0,0.0);
  }
  return fg;
}


complex structure_factor_ext_tbl(
  HKL g, int num_sym_op, SYM_OP sym_op[],
  int n_atom, ATOM atm[], XDT xdyl[], ATOMF_BK ftbl[],
  double kvolts, CELL_PAR bv, char *lattice_type,
  int *ct_fg_table, int g_table_size, HKL *g_table, complex *fg_table
  )
{
  complex fg;
  int ng;
  
  if ( ( ng = search_g_table( *ct_fg_table, g, g_table) ) >= 0 ){	/* found */
    fg = fg_table[ng];
  }
  else{ /* not found in the table */
    fg = structure_factor_ext(
      g, num_sym_op, sym_op, n_atom, atm, xdyl, ftbl, kvolts, bv, lattice_type
      );
    stock_fg_tbl_sym(
      g, fg, ct_fg_table, g_table_size, g_table, fg_table, num_sym_op, sym_op
      );
  }
  return fg;
}


complex calc_z_factor(
  HKL g, int num_sym_op, SYM_OP sym_op[],
  int n_atom, ATOM atm[],	XDT xdyl[], double kvolts,
  CELL_PAR bv, char *lattice_type
  )
{
  /*
   * Calc. of "Z" structure factor
   * \Sigma Z_i exp(-2pi*i*B*s^2) exp(-2pi*i*g*r_i)
   */
  int  i;
  double  *f_ela, *f_inela, sq_s;
  complex  ZFg;
  
  /*  s = sin(theta) / lambda = 1 / (2d) = g / 2  */
  sq_s = sq_abs( hkl2rhkl(g), bv) / 4.0;
  f_ela = new_vector(double, n_atom);  chk_vector(f_ela);
  f_inela = new_vector(double, n_atom);  chk_vector(f_inela);
  for(i=0; i<n_atom; i++){
    bij_to_b( g, sq_s, &(atm[i]), bv );
    f_ela[i]  = xdyl[i].znum;  /* Z number */
    f_inela[i]  = 0.0;
  }
  ZFg = calc_structure_factor_ph(
    g, sq_s, num_sym_op, sym_op, n_atom, atm, f_ela, f_inela
    );
  free_vector( f_ela );
  free_vector( f_inela );
  return ZFg;
}


/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */
complex calc_Fg_ela(
  HKL g, int num_sym_op, SYM_OP sym_op[],
  int n_atom, ATOM atm[], XDT xdyl[], ATOMF_BK ftbl[],
  double kvolts, CELL_PAR bv
  )
     /*
      * Calc. of only the elastic part of structure factor F_g [A] (rf version)
      *                   ^^^^^^^^^^^^
      *   using Ichimiya's formula or Bird & King's func.
      * for lofg_to_fg()
      */
{
  int i;
  double *f_ela, *f_inela, sq_s, s;
  complex Fg, atomf;
  
  /*    s = sin(theta) / lambda = 1 / (2d) = g / 2   */
  sq_s = sq_abs( hkl2rhkl(g), bv) / 4.0;
  s = sqrt(sq_s);
  
  f_ela = new_vector(double, n_atom);		chk_vector(f_ela);
  f_inela = new_vector(double, n_atom);	chk_vector(f_inela);
  
  for(i=0; i<n_atom; i++){
    /* cst_aniso_bij(&(atm[i])); */			/* Constraint for anisotropic bij */
    
    if(atm[i].af == 0){	/* X-ray's f and Mott formula */
      f_ela[i]	= atomsf_xdt( sq_s, xdyl[i] );
      f_inela[i]	= 0.0;
    }
    else if(atm[i].af == 1 || atm[i].af == 2){	/* Bird & King */
      atom_bk( ftbl[i], s, atm[i].b, &atomf, kvolts );
      f_ela[i]	= atomf.re;
      f_inela[i]	= 0.0;
    }
    else{
      printf("$$ Invalid af. Abort.\n");	exit(EXIT_FAILURE);
    }
  }
  
  Fg = calc_structure_factor_ph( g, sq_s, num_sym_op, sym_op, n_atom, atm, f_ela, f_inela );
  free_vector( f_ela );	free_vector( f_inela );
  return Fg;
}


complex calc_Fg_inela(
  HKL g, int num_sym_op, SYM_OP sym_op[],
  int n_atom, ATOM atm[], ATOMF_BK ftbl[], double kvolts, CELL_PAR bv
  )
     /*
      * Calc. of only the inelastic part of structure factor F_g [A] (rf version)
      *                   ^^^^^^^^^^^^^^
      *   using Ichimiya's formula or Bird & King's func.
      * for lofg_to_fg()
      */
{
  int i;
  double *f_ela, *f_inela, sq_s, s;
  complex Fg, atomf;
  
  /*    s = sin(theta) / lambda = 1 / (2d) = g / 2   */
  sq_s = sq_abs( hkl2rhkl(g), bv) / 4.0;
  s = sqrt(sq_s);
  
  f_ela = new_vector(double, n_atom);		chk_vector(f_ela);
  f_inela = new_vector(double, n_atom);	chk_vector(f_inela);
  
  for(i=0; i<n_atom; i++){
    if(atm[i].af == 0 || atm[i].af == 1){	/* Bird & King */
      atom_bk( ftbl[i], s, atm[i].b, &atomf, kvolts );
      f_ela[i]	= 0.0;
      f_inela[i]	= atomf.im;
    }
    else{
      printf("$$ Invalid af. Abort.\n");	exit(EXIT_FAILURE);
    }
  }
  
  Fg = calc_structure_factor_ph( g, sq_s, num_sym_op, sym_op, n_atom, atm, f_ela, f_inela );
  free_vector( f_ela );	free_vector( f_inela );
  return Fg;
}


complex calc_structure_factor_f_out(
  FILE *fo, HKL g, int num_sym_op, SYM_OP sym_op[],
  int n_atom, ATOM atm[], XDT xdyl[], ATOMF_BK ftbl[],
  double kvolts, CELL_PAR bv
  )
{
  int i;
  double *f_ela, *f_inela, sq_s, s;
  complex Fg, Fg_ela, Fg_inela, atomf;

  /*    s = sin(theta) / lambda = 1 / (2d) = g / 2   */
  sq_s = sq_abs( hkl2rhkl(g), bv) / 4.0;
  s = sqrt(sq_s);

  f_ela = new_vector(double, n_atom);		chk_vector(f_ela);
  f_inela = new_vector(double, n_atom);	chk_vector(f_inela);

  fprintf(fo, " g = %d %d %d\n", g.h, g.k, g.l);

  /* ----- Elastic part of Fg ----- */
  for(i=0; i<n_atom; i++){
    /* cst_aniso_bij(&(atm[i])); */ /* Constraint for anisotropic bij */
    if(atm[i].af == 0){	/* Mott formula + Ichimiya's absorption */
      bij_to_b( g, sq_s, &(atm[i]), bv );
      f_ela[i]	= atom_scat_f_dtx( sq_s, xdyl[i] );				/* <<<<< */
      f_inela[i]	= 0.0;
    }
    else if(atm[i].af == 1 || atm[i].af == 2){	/* Bird & King */
      bij_to_b( g, sq_s, &(atm[i]), bv );
      atom_bk( ftbl[i], s, atm[i].b, &atomf, kvolts );
      f_ela[i]	= atomf.re;									/* <<<<< */
      f_inela[i]	= 0.0;
    }
    else{
      printf("$$ Invalid af. Abort.\n");	exit(EXIT_FAILURE);
    }
  }
  Fg_ela = calc_structure_factor_ph(
    g, sq_s, num_sym_op, sym_op, n_atom, atm, f_ela, f_inela
    );
  
  /* ----- Inelastic part of Fg ----- */
  for(i=0; i<n_atom; i++){
    if(atm[i].af == 0){	/* Mott formula + inela part of BK */
      bij_to_b( g, sq_s, &(atm[i]), bv );
      atom_bk( ftbl[i], s, atm[i].b, &atomf, kvolts );
      f_ela[i]	= 0.0;
      f_inela[i]	= atomf.im;
    }
    else if(atm[i].af == 1){	/* Bird & King */
      bij_to_b( g, sq_s, &(atm[i]), bv );
      atom_bk( ftbl[i], s, atm[i].b, &atomf, kvolts );
      f_ela[i]	= 0.0;
      f_inela[i]	= atomf.im;
    }
    else{
      printf("$$ Invalid af. Abort.\n");	exit(EXIT_FAILURE);
    }
  }
  Fg_inela = calc_structure_factor_ph(
    g, sq_s, num_sym_op, sym_op, n_atom, atm, f_ela, f_inela
    );
  
  /* ----- Elastic and inelastic part of Fg ----- */
  fprintf(fo, "   s = g / 2 = %lf [A^{-1}]\n", s);
  fprintf(fo, "   atom\telement\telastic_f\tinelastic_f [A]\n");
  for(i=0; i<n_atom; i++){
    if(atm[i].af == 0){	/* Mott formula + Inela BK */
      bij_to_b( g, sq_s, &(atm[i]), bv );
      atom_bk( ftbl[i], s, atm[i].b, &atomf, kvolts );
      f_ela[i]	= atom_scat_f_dtx( sq_s, xdyl[i] );
      f_inela[i]	= atomf.im;
    }
    else if(atm[i].af == 1){	/* Bird & King */
      bij_to_b( g, sq_s, &(atm[i]), bv );
      atom_bk( ftbl[i], s, atm[i].b, &atomf, kvolts );
      f_ela[i]	= atomf.re;
      f_inela[i]	= atomf.im;
    }
    else{
      printf("$$ Invalid af. Abort.\n");	exit(EXIT_FAILURE);
    }
    fprintf(fo, "   %d\t\t%s\t\t%lf\t%lf\n", i, atm[i].element, f_ela[i], f_inela[i]);
    fprintf(fo, "   %d\t\t%s\t\t%lf\t%lf\t%lf\n", i, atm[i].element, f_ela[i], f_inela[i], atm[i].b);
  }
  Fg = calc_structure_factor_ph( g, sq_s, num_sym_op, sym_op, n_atom, atm, f_ela, f_inela );
  fprintf(fo, "  Fg = (%lf,%lf)\n", Fg.re, Fg.im);
  fprintf(fo, "   (Elastic part   = (%lf,%lf) )\n", Fg_ela.re, Fg_ela.im);
  fprintf(fo, "   (Inelastic part = (%lf,%lf) )\n", Fg_inela.re, Fg_inela.im);
  if (c_abs(Fg_inela) > epsilon.eps){
    fprintf(fo, "$$   (abs(ela/inela) = %lf)\n", c_abs(Fg_ela) / c_abs(Fg_inela) );
  }
  else{
    ;
  }
  free_vector( f_ela );	free_vector( f_inela );
  return Fg;
}


/* ----- */

complex lofg_to_fg(
  HKL g, int lof_type, LO_FG *Fg_lo, int num_sym_op, SYM_OP sym_op[],
  int n_atom, ATOM atm[], ATOMF_BK ftbl[], XDT *xdyl, double kvolts, CELL_PAR bv
  )
     /*
      * Low-order structure factor
      */
{
  double sq_s;
  complex Fg, Fg_ela, Fg_inela;
  
  if(lof_type == 1){		/* type 1: whole part of fg */
    Fg.re = Fg_lo->re;
    Fg.im = Fg_lo->im;
  }
  else if(lof_type == 2){	/* type 2: only elastic part of fg */
    Fg_ela.re = Fg_lo->ela_re;
    Fg_ela.im = Fg_lo->ela_im;
    Fg_inela = calc_Fg_inela( g, num_sym_op, sym_op, n_atom, atm, ftbl, kvolts, bv );
    Fg = c_add( Fg_ela, Fg_inela );
  }
  else if(lof_type == 3){	/* type 3: f_ela and f_inela */
    sq_s = sq_abs( hkl2rhkl(g), bv) / 4.0;
    Fg = calc_structure_factor_ph(
      g, sq_s, num_sym_op, sym_op, n_atom, atm, Fg_lo->fe, Fg_lo->fi
      );
  }
  else{
    printf("$$ Invalid lof_type. Abort.\n");	exit(EXIT_FAILURE);
  }
  /*printf("$$ Low order Fg = %lf %lf\n", Fg.re, Fg.im);*/
  return Fg;
}

/* ----- */

complex calc_Xray_Fg_ela(
  HKL g, int num_sym_op, SYM_OP sym_op[],
  int n_atom, ATOM atm[], XDT xdyl[], double kvolts, CELL_PAR bv
  )
     /*
      * Calc. of only the elastic part of X-ray structure factor F_g_Xray [A] (rf version)
      *                   ^^^^^^^^^^^^    ^^^^^^^^^^^^^^^^^^^^^^^
      *   using Ichimiya's formula or Bird & King's func.
      * for lofg_to_fg()
      */
{
  int i;
  double *f_ela, *f_inela, sq_s, s;
  complex Fgx;
  
  /*    s = sin(theta) / lambda = 1 / (2d) = g / 2   */
  sq_s = sq_abs( hkl2rhkl(g), bv) / 4.0;
  s = sqrt(sq_s);
  
  f_ela = new_vector(double, n_atom);		chk_vector(f_ela);
  f_inela = new_vector(double, n_atom);	chk_vector(f_inela);
  
  for(i=0; i<n_atom; i++){
    /* cst_aniso_bij(&(atm[i])); */			/* Constraint for anisotropic bij */
    f_ela[i]	= xray_asf( sq_s, xdyl[i] );	/* X-ray's f */
    f_inela[i]	= 0.0;
  }
  
  Fgx = calc_structure_factor_ph( g, sq_s, num_sym_op, sym_op, n_atom, atm, f_ela, f_inela );
  free_vector( f_ela );	free_vector( f_inela );
  return Fgx;
}
